
#ifndef _LINUX_SWSUSP_DEBUG_H
#define _LINUX_SWSUSP_DEBUG_H

#include <linux/suspend.h>
#include <linux/suspend-version-specific.h>

/* Solely for comparison with the version specific patch revision */
#define SWSUSP_CORE_REVISION 0x201

#define name_suspend "SoftwareSuspend2: "
#define swsusp_version "2.0-rc4"
#define name_swsusp  "Swsusp " swsusp_version ": "
#define console_suspend " S U S P E N D   T O   D I S K " /* Same length to ensure one overwrites the other */
#define console_resume  "R E S U M E   F R O M   D I S K"

#define TEST_RESULT_STATE(bit) (test_bit(bit, &swsusp_result))
#define SET_RESULT_STATE(bit) (test_and_set_bit(bit, &swsusp_result))
#define CLEAR_RESULT_STATE(bit) (test_and_clear_bit(bit, &swsusp_result))

#define TEST_ACTION_STATE(bit) (test_bit(bit, &swsusp_action))
#define SET_ACTION_STATE(bit) (test_and_set_bit(bit, &swsusp_action))
#define CLEAR_ACTION_STATE(bit) (test_and_clear_bit(bit, &swsusp_action))

#define TEST_DEBUG_STATE(bit) (test_bit(bit, &swsusp_debug_state))
#define SET_DEBUG_STATE(bit) (test_and_set_bit(bit, &swsusp_debug_state))
#define CLEAR_DEBUG_STATE(bit) (test_and_clear_bit(bit, &swsusp_debug_state))

extern int suspend_console;

/* first status register - this is swsusp's return code. */
#define SUSPEND_ABORTED			0
#define SUSPEND_ABORT_REQUESTED		1
#define SUSPEND_NOSTORAGE_AVAILABLE	2
#define SUSPEND_INSUFFICIENT_STORAGE	3
#define SUSPEND_FREEZING_FAILED		4
#define SUSPEND_UNEXPECTED_ALLOC	5
#define SUSPEND_KEPT_IMAGE		6
#define SUSPEND_WOULD_EAT_MEMORY	7
#define SUSPEND_UNABLE_TO_FREE_ENOUGH_MEMORY 8

/* second status register */
#define SUSPEND_REBOOT		0
#define SUSPEND_NO_OUTPUT	1
#define SUSPEND_PAUSE		2
#define SUSPEND_SLOW		3
#define SUSPEND_BEEP		4
#define SUSPEND_NOPAGESET2	7
#define SUSPEND_LOGALL		8
/* Set to disable compression when compiled in */
#define SUSPEND_NO_COMPRESSION	9
//#define SUSPEND_ENABLE_KDB	10
#define SUSPEND_CAN_CANCEL	11
#define SUSPEND_KEEP_IMAGE	13
#define SUSPEND_FREEZER_TEST	14
#define SUSPEND_FREEZER_TEST_SHOWALL 15

/* debug sections  - if debugging compiled in */
#define SUSPEND_ANY_SECTION	0
#define SUSPEND_FREEZER		1
#define SUSPEND_EAT_MEMORY 	2
#define SUSPEND_PAGESETS	3
#define SUSPEND_IO		4
#define SUSPEND_BMAP		5
#define SUSPEND_SWAP		9
#define SUSPEND_MEMORY		10
#define SUSPEND_RANGES		11
#define SUSPEND_SPINLOCKS	12
#define SUSPEND_MEM_POOL	13
#define SUSPEND_RANGE_PARANOIA	14
/* debugging levels - if debugging compiled in.
 * These are straight integer values, not bit flags.
 * (Stored as LEVEL << 16 + section_flags).
 */
#define SUSPEND_ERROR		2
#define SUSPEND_LOW	 	3
#define SUSPEND_MEDIUM	 	4
#define SUSPEND_HIGH	  	5
#define SUSPEND_VERBOSE		6

/* 
 *   #defs that are dependant on CONFIG_SOFTWARE_SUSPEND_DEBUG
 *
 */
#if defined(CONFIG_SOFTWARE_SUSPEND_DEBUG)
void printnolog(int mask, int level, int restartline, const char *fmt, ...);
void printlog(int mask, int level, const char *fmt, ...);

#define MDELAY(a)	if (TEST_ACTION_STATE(SUSPEND_SLOW)) { mdelay(a); } else { do { } while (0); }

extern int currentbeep;
#define beepOK \
	if (TEST_ACTION_STATE(SUSPEND_BEEP)) { \
		currentbeep += 200; \
		kd_mksound(currentbeep,HZ/8); \
		mdelay(150); \
	} else { \
		do { } while (0); \
	}
#define beepERR \
	if (TEST_ACTION_STATE(SUSPEND_BEEP)) { \
		kd_mksound(300,HZ/4); \
		mdelay(300); \
	} else { \
		do { } while (0); \
	} 

#else // #ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define MDELAY(a) do { } while (0)
#define beepOK do { } while (0)
#define beepERR do { } while (0)
#define printlog(...) do { } while(0)
#define printnolog(...) do { } while(0)
#endif // #ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG

#endif
